# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  genemail.email
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/09
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

from __future__ import absolute_import

import re, copy, base64, mimetypes
import xml.etree.ElementTree as ET
import html2text
from templatealchemy.util import adict
import email.Encoders, email.Message, email.MIMEMultipart, email.MIMEText
import email.MIMEImage, email.Utils

from . import util
from .idict import idict

#------------------------------------------------------------------------------
__all__ = ('Email',)
xmlns = 'http://pythonhosted.org/genemail/xmlns/1.0'
htmlns = util.htmlns
class MissingHeader(Exception): pass

#------------------------------------------------------------------------------
class Email(object):

  DEFAULTS = dict(
    structure = {

      #  default mail structure:
      #
      #     multipart/mixed [IFF there are attachments without "Content-ID"]
      #     |-- multipart/alternative
      #     |   |-- text/plain
      #     |   `-- multipart/related; type="text/html"
      #     |       |-- text/html
      #     |       `-- image/png... [attachments with "Content-ID"]
      #     `-- application/octet-stream... [attachments without "Content-ID"]
      #
      #               Content-Type: image/png
      #               Content-Transfer-Encoding: base64
      #               Content-Location: file:///.../...
      #               Content-ID: <ImageName>

      'mime:mixed; optimize=True': [
        {'mime:alternative': [
          'email:text',
          {'mime:related; type="text/html"': [
            'email:html',
            {'mime:attachments': 'email:attachments; cid=True'},
            ]},
          ]},
        {'mime:attachments': 'email:attachments; cid=False'},
        ]

      },
    params               = {},
    attachments          = [],
    headers              = idict(),
    includeComponents    = [],
    maxSubjectLength     = 512,
    snipIndicator        = '[...]',
    textEncoding         = None,
    htmlEncoding         = None,
    transferEncoding     = None,
    encoding             = None,
    boundary             = None,
    minimalMime          = True,
    )

  #----------------------------------------------------------------------------
  def __init__(self, manager, name, provider=None, default=None):
    self.manager  = manager
    self.name     = name
    self.provider = provider or self.manager.provider
    self.template = self.provider.getTemplate(self.name)
    self._headers = idict()
    if default is None:
      default = adict(self.DEFAULTS)
    for attr in self.DEFAULTS.keys():
      setattr(self, attr, copy.deepcopy(getattr(default, attr, None)))

    # self.attachmentTable    = None
    # for key, val in self.provider.getMap('headers', {}).items():
    #   self.headers[key] = val

  #----------------------------------------------------------------------------
  def __setitem__(self, key, val): self.params[key] = val
  def __getitem__(self, key): return self.params[key]
  def __delitem__(self, key): del self.params[key]

  #----------------------------------------------------------------------------
  def setHeader(self, key, val):
    'Set the email header `key` to `val`, overriding all other header sources.'
    self.headers[key] = val

  #----------------------------------------------------------------------------
  def getHeader(self, key):
    '''
    Get the current value of the header `key`. Note that this will only return
    the value as set with a prior :meth:`setHeader` call - it will *not*
    extract headers set by the template(s).
    '''
    return self.headers[key]

  #----------------------------------------------------------------------------
  def hasHeader(self, key):
    '''
    Returns whether or not the header `key` has been set with a prior call
    to :meth:`setHeader` - it will *not* extract headers set by the template(s).
    '''
    return self.headers.has_key(key)

  #----------------------------------------------------------------------------
  def delHeader(self, key):
    '''
    If the header `key` had been set with a prior call to
    :meth:`setHeader`, this will unset the header. Note that it will
    *not* delete any headers set by the template(s).
    '''
    del self.headers[key]

  #----------------------------------------------------------------------------
  # note: for some reason @property...(setter|deleter) don't work with
  # with setattr(...) in python 2.7.3. ugh. need to use old-style.
  def getHeaders(self):
    return self._headers
  def setHeaders(self, val):
    self._headers = idict(val)
  def delHeaders(self):
    self._headers = idict()
  headers = property(getHeaders, setHeaders, delHeaders)

  #----------------------------------------------------------------------------
  def getSettings(self):
    '''
    Returns a dictionary of settings set by the template (this is so
    that templates can communicate non-output parameters back to the
    calling application).
    '''
    if self.template.meta.settings:
      return self.template.meta.settings
    return adict()

  #----------------------------------------------------------------------------
  def getSetting(self, key, default=None):
    '''
    Returns the template-set setting name `key`. If not set, `default`
    is returned. See :meth:`getSettings` for details.
    '''
    return self.getSettings().get(key, default)

  #----------------------------------------------------------------------------
  def addAttachment(self, name, value, cid=False, contentType=None):
    '''
    Add an attachment to the email. The `name` is the default filename that
    will be offered when the recipient tries to "Save..." the attachment. The
    `value` is the actual content of the attachment. If `cid` is True, then
    the attachment will be stored as an embedded object, and will therefore
    not be directly saveable by the recipient and it can be accessed from
    within the HTML, for example, with the `name` set to ``'logo.png'``, the
    following will result in a valid image reference in the HTML::

      <img alt="The Logo" src="cid:logo.png"/>

    (note the ``cid:`` prefix that must be added in the HTML.)
    '''
    self.attachments.append(
      adict(name=name, value=value, cid=cid, contentType=contentType))

  #----------------------------------------------------------------------------
  def getTemplateXml(self, fallbackToNone=True):
    params = dict(self.params)
    params.update({'genemail_format': 'xml'})
    for fmt in ('xml', 'xhtml', 'html'):
      if fmt in self.template.meta.formats:
        return util.parseXml(self.template.render(fmt, params))
    if fallbackToNone:
      try:
        return util.parseXml(self.template.render(None, params))
      except ET.ParseError:
        return None
    return None

  #----------------------------------------------------------------------------
  def getTemplateHeaders(self):
    ret = idict()
    # todo: this requires that headers be defined as separate elements, eg:
    #         <email:header name="..." value="...">...</email:header>
    #       rather than being able to tag an existing node, such as:
    #         <p>This email was sent to <span email:header="To">...</span>.</p>
    xdoc = self.getTemplateXml()
    if xdoc is None:
      return ret
    etag = '{%s}header' % (xmlns,)
    for node in xdoc.iter():
      if node.tag == etag:
        ret[node.get('name')] = node.get('value') or node.text
        continue
      if node.get(etag) is not None:
        ret[node.get(etag)] = node.text
    return ret

  #----------------------------------------------------------------------------
  def getTemplateAttachments(self):
    atts = {}
    for att in self.template.meta.attachments or []:
      atts[att.name] = att
    xdoc = self.getTemplateXml()
    if xdoc is None:
      return atts.values()
    for node in xdoc.iter('{%s}attachment' % (xmlns,)):
      att = adict(
        name        = node.get('name'),
        contentType = node.get('content-type', None),
        value       = node.get('value', None) or node.text,
        cid         = node.get('cid', 'false').lower() == 'true',
        )
      if node.get('encoding', None) == 'base64':
        att.value = base64.b64decode(att.value)
      atts[att.name] = att
    return atts.values()

  #----------------------------------------------------------------------------
  def getTemplateStyle(self):
    ret  = []
    if 'css' in self.template.meta.formats:
      params = dict(self.params)
      params.update({'genemail_format': 'css'})
      ret.append(self.template.render('css', params))
    xdoc = self.getTemplateXml()
    if xdoc is None:
      return ' '.join(ret)
    # todo: what if the node has a "src" attribute...
    for node in xdoc.findall('{%s}head/{%s}style[@type="text/css"]' % (htmlns, htmlns)):
      ret.append(node.text)
    return ' '.join(ret)

  #----------------------------------------------------------------------------
  def getHtml(self, standalone=False, extraparams=None):
    '''
    Returns the raw HTML that would be generated if the email were
    sent with the current settings. Use :meth:`send` to actually send
    the email.
    '''

    # todo: what if 'html' is not in `includeComponents`?

    if standalone:
      ret = self.getHtml(extraparams=extraparams)
      return self.inlineCidAttachments(ret)

    params = dict(self.params)
    params.update({'genemail_format': 'html'})
    if extraparams:
      params.update(extraparams)

    # todo: try alternative formats if 'html' isn't available? eg xhtml...
    html = self.template.render('html', params)

    # todo: this double roundtrip of parse/serialize html is ridiculous

    # remove all genemail xmlns elements and attributes and non-inline css
    # note: using list() so that i can mutate the underlying object
    html = util.parseXml(html)
    def stripSpecial(el):
      for attr in list(el.keys()):
        if attr.startswith('{%s}' % (xmlns,)):
          del el.attrib[attr]
      for node in list(el.getchildren()):
        if node.tag.startswith('{%s}' % (xmlns,)):
          el.remove(node)
          continue
        stripSpecial(node)
    stripSpecial(html)
    for topnode in html.getchildren():
      if topnode.tag == '{%s}head' % (htmlns,):
        for node in list(topnode.getchildren()):
          if node.tag == '{%s}style' % (htmlns,) and node.get('type') == 'text/css':
            topnode.remove(node)

    style = (self.getTemplateStyle() or '').strip()
    if len(style) > 0:
      html = util.inlineHtmlStyling(html, style)

    return util.serializeHtml(html)

  #----------------------------------------------------------------------------
  def inlineCidAttachments(self, html):
    for att in self.getAttachments():
      if not att.cid:
        continue
      ct = att.contentType or 'application/octet-stream'
      value = 'data:' + ct + ';base64,' + base64.b64encode(att['value'])
      # todo: this is a "brute-force" approach... i should replace only
      #       attributes that have this exact value...
      html = html.replace('cid:' + att.name, value)
    return html

  #----------------------------------------------------------------------------
  def getText(self):
    '''
    Returns the plain-text version of the email that would be
    generated if it were sent with the current settings. Use
    :meth:`send` to actually send the email.
    '''
    # todo: what if 'text' is not in `includeComponents`?
    params = dict(self.params)
    params.update({'genemail_format': 'text'})
    if 'text' in self.template.meta.formats:
      return self.template.render('text', params)
    try:
      html = self.getHtml(extraparams={'genemail_format': 'text'})
    except Exception:
      return self.template.render(None, params)
    # todo: it would be interesting to be able to configure html2text to only
    #       put it footnotes for IMG tags that had non-"cid:" image references...
    text = html2text.html2text(html)
    # TODO: html2text should have taken care of this... but since it hasn't,
    #       are there any other characters that need to be escaped?...
    # todo: this seems a bit brute-force to reduce to ascii? perhaps
    #       the caller should inspect the text and change the content type?
    #       or, better yet, have this return a tuple with the content type
    return util.removeCids(util.reduce2ascii(text)).strip() + '\n'

  subject_collapse_spaces = re.compile(r'[\s]+', re.DOTALL)
  subject_remove_nonascii = re.compile(r'[^ -~]+', re.DOTALL)

  #----------------------------------------------------------------------------
  def _cleanSubject(self, subject):
    ret = self.subject_collapse_spaces.sub(' ', subject)
    ret = self.subject_remove_nonascii.sub('', ret)
    ret = util.reduce2ascii(ret.strip())
    if self.maxSubjectLength is not None and self.snipIndicator is not None:
      if len(ret) > self.maxSubjectLength:
        ret = ret[:self.maxSubjectLength - len(self.snipIndicator)] + self.snipIndicator
    return ret

  #----------------------------------------------------------------------------
  def getSubject(self):
    '''
    Returns the ``Subject`` header of the email that would be
    generated if it were sent with the current settings. Use
    :meth:`send` to actually send the email.
    '''
    if 'subject' in self.headers:
      return self.headers['subject']
    if 'subject' in self.template.meta.formats:
      params = dict(self.params)
      params.update({'genemail_format': 'subject'})
      # todo: check subject encoding?...
      return self.template.render('subject', self.params)
    # todo: what if there is no XML format?...
    # note: purposefully NOT using encoding=self.encoding so that i can
    #       then more cleanly replace HTML entity characters...
    # todo: clean up this entity-replacement strategy
    xdoc = self.getTemplateXml()
    if xdoc is None:
      return self._cleanSubject(self.getText())
    etag = '{%s}subject' % (xmlns,)
    ret = []
    for node in xdoc.iter():
      if node.tag == etag or node.get(etag) == 'content':
        ret.append(node.text)
    ret = ' '.join(ret).encode('us-ascii', 'genemail_unicode2ascii')
    ret = self._cleanSubject(ret)
    if ret:
      return ret
    return self._cleanSubject(self.getText())

  #----------------------------------------------------------------------------
  def getAttachments(self):
    atts = {att.name: att for att in self.getTemplateAttachments()}
    for att in self.attachments:
      atts[att.name] = att
    for att in atts.values():
      if att.contentType:
        continue
      att.contentType = mimetypes.guess_type(att.name or '', False)[0] \
        or 'application/octet-stream'
    return atts.values()

  #----------------------------------------------------------------------------
  def getOutputHeaders(self):
    curheaders = idict()
    curheaders.update(self.getTemplateHeaders())
    curheaders.update(self.headers)
    defaultHeaders = {
      'Subject' : lambda: self.getSubject(),
      'Date'    : lambda: email.Utils.formatdate(),
      }
    for name, value in defaultHeaders.items():
      if name not in curheaders:
        curheaders[name] = value() if callable(value) else value
    if hasattr(self.manager, 'updateHeaders'):
      self.manager.updateHeaders(self, curheaders)
    return curheaders

  #----------------------------------------------------------------------------
  def getSmtpData(self):
    '''
    Returns the raw SMTP data that would be sent if the email were sent
    with the current settings. Use :meth:`send` to actually send the email.
    '''
    return self._getMimeMessage(self.getOutputHeaders()).as_string()

  #----------------------------------------------------------------------------
  def _getMimeMessage(self, curheaders):

    curheaders = idict(curheaders)

    # reduce headers to those that should be in outbound email
    # todo: should i instead be filtering for known allowed headers instead
    #       of removing known disallowed headers?...
    for header in ('bcc',):
      if header in curheaders:
        del curheaders[header]

    # the following text/html encoding selection is done only because
    # utf-8 results in base64 transfer encoding, which sucks because it
    # is not human-readable... hence trying others first. and the problem
    # with iso-8859-1 is that the "&copy;" and "&nbsp;" entities don't
    # show up, but it uses quoted-printable...
    # ==> UPDATE: i can't seem to reproduce this... the following translations
    #             occur: &mdash; => &#8212;, &copy; => &#169;, &nbsp; => &#160;

    # TODO: address the other self.encoding uses and see if the default
    #       should really be us-ascii instead of None...

    funcs = adict()

    def make_email_text():
      if self.includeComponents and 'text' not in self.includeComponents:
        return None
      txtenc  = self.textEncoding or self.transferEncoding or self.encoding
      if txtenc is not None:
        return email.MIMEText.MIMEText(self.getText(), 'plain', txtenc)
      # do several encoding round trips until one is found that does not
      # degrade the content
      for txtenc in [ 'ascii', 'iso-8859-1', 'utf-8' ]:
        src = self.getText()
        try: txt = email.MIMEText.MIMEText(src, 'plain', txtenc)
        except UnicodeEncodeError: continue
        if src == txt.get_payload(decode=True):
          return txt
      raise TypeError('could not encode email text component - try setting an encoding')
    funcs.make_email_text = make_email_text

    def make_email_html():
      if self.includeComponents and 'html' not in self.includeComponents:
        return None
      htmlenc = self.htmlEncoding or self.transferEncoding or self.encoding
      if htmlenc is not None:
        return email.MIMEText.MIMEText(self.getHtml(), 'html', htmlenc)
      # do several encoding round trips until one is found that does not
      # degrade the content
      for htmlenc in [ 'ascii', 'iso-8859-1', 'utf-8' ]:
        src = self.getHtml()
        try: html = email.MIMEText.MIMEText(src, 'html', htmlenc)
        except UnicodeEncodeError: continue
        if src == html.get_payload(decode=True):
          return html
      raise TypeError('could not encode email html component - try setting an encoding')
    funcs.make_email_html = make_email_html

    funcs.atts_cache = None
    def make_email_attachments(cid=False):
      if self.includeComponents and 'attachments' not in self.includeComponents:
        return None
      if funcs.atts_cache is None:
        funcs.atts_cache = self.getAttachments()
      return [att for att in funcs.atts_cache if att.cid == cid]
    funcs.make_email_attachments = make_email_attachments

    funcs.bndidx = 0
    def make_boundary(type):
      if not self.boundary:
        return None
      funcs.bndidx += 1
      return '==%s-%s-%i==' % (self.boundary, type, funcs.bndidx)
    funcs.make_boundary = make_boundary

    def comp_extend(comp, spec):
      if isinstance(spec, basestring):
        spec = [spec]
      for item in ( spec or [] ):
        item = funcs.make_component(item)
        if not item:
          continue
        if isinstance(comp, list):
          if isinstance(item, list):
            comp.extend(item)
          else:
            comp.append(item)
        else:
          if isinstance(item, list):
            for cur in item:
              comp.attach(cur)
          else:
            comp.attach(item)
      return comp
    funcs.comp_extend = comp_extend

    def checkMinimalMime(comp, optimize=None):
      if optimize is True or self.minimalMime:
        load = comp.get_payload()
        if not isinstance(load, basestring) \
            and len(load) == 1:
          return load[0]
      return comp

    def make_mime_mixed(spec, optimize=None):
      comp = email.MIMEMultipart.MIMEMultipart(
        'mixed', boundary=make_boundary('mix'))
      return checkMinimalMime(funcs.comp_extend(comp, spec))
    funcs.make_mime_mixed = make_mime_mixed

    def make_mime_alternative(spec, optimize=None):
      comp = email.MIMEMultipart.MIMEMultipart(
        'alternative', boundary=make_boundary('alt'))
      return checkMinimalMime(funcs.comp_extend(comp, spec))
    funcs.make_mime_alternative = make_mime_alternative

    def make_mime_related(spec, type=None, optimize=None):
      # TODO: do something with `type`...
      comp = email.MIMEMultipart.MIMEMultipart(
        'related', boundary=make_boundary('rel'))
      return checkMinimalMime(funcs.comp_extend(comp, spec))
    funcs.make_mime_related = make_mime_related

    def make_mime_attachments(spec):
      if self.includeComponents and 'attachments' not in self.includeComponents:
        return None
      atts = funcs.comp_extend([], spec)
      if not atts:
        return None
      ret = []
      for att in atts:
        maintype, subtype = (att.contentType or 'application/octet-stream').split('/', 1)
        if maintype == 'text':
          # note: we should handle calculating the charset
          matt = email.MIMEText.MIMEText(att.value, _subtype=subtype)
        elif maintype == 'image':
          matt = email.MIMEImage.MIMEImage(att.value, _subtype=subtype, name=att.name)
        elif maintype == 'audio':
          matt = email.MIMEAudio.MIMEAudio(att.value, _subtype=subtype)
        else:
          matt = email.MIMEBase.MIMEBase(maintype, subtype)
          matt.set_payload(att.value)
          email.Encoders.encode_base64(matt)
        if att.cid:
          matt.add_header('Content-Disposition', 'attachment')
          matt.add_header('Content-ID', '<' + att.name + '>')
        else:
          matt.add_header('Content-Disposition', 'attachment', filename=att.name)
        ret.append(matt)
      return ret
    funcs.make_mime_attachments = make_mime_attachments

    def make_component(spec):
      kls, comp, kw = spec, None, dict()
      if isinstance(spec, dict) and len(spec) == 1:
        kls = spec.keys()[0]
      if isinstance(kls, basestring) and ':' in kls:
        kls, comp = kls.split(':', 1)
        if ';' in comp:
          comp, kw = comp.split(';', 1)
          kw = [pairs.strip().split('=', 1)
                for pairs in kw.strip().split(',')]
          kw = {k: eval(v) for k, v in kw}
      if kls == 'email':
        return funcs['make_email_' + comp](**kw)
      if kls == 'mime':
        return funcs['make_mime_' + comp](spec.values()[0], **kw)
      raise SyntaxError('unsupported email structure component: %r' % (spec,))
    funcs.make_component = make_component

    if self.includeComponents == ['text']:
      msg = make_email_text()
    else:
      msg = make_component(self.structure)

    for k,v in curheaders.items():
      msg[util.smtpHeaderFormat(k)] = v

    return msg

  #----------------------------------------------------------------------------
  def send(self, mailfrom=None, recipients=None):
    '''
    Send the email. The ``To`` header can be overriden with the `recipients`
    parameter, and the ``From`` header con be overriden with the `mailfrom`
    parameter. Note that if these parameters are used, then the headers are
    NOT adjusted to reflect the new settings (this is to allow ``Bcc`` type
    behavior).

    :Parameters:

    mailfrom : str, optional

      Email address to set as the ``MAIL FROM`` address on the SMTP
      protocol level. Note that this will NOT set or override the
      active ``From`` header. If not specified, the address will be
      extracted from the ``From`` header.

    recipients : { str, list(str) }, optional

      Email addresses to send this email to, on the SMTP protocol
      level. Note that this will NOT set or override the active ``To``
      header. If not specified, the addresses will be extracted from
      the ``To``, ``CC``, and ``BCC`` headers.
    '''
    hdrs = self.getOutputHeaders()
    if mailfrom is None:
      mailfrom = util.extractEmails(hdrs.get('from'))
      if not mailfrom:
        raise MissingHeader('email source ("from") not specified')
      mailfrom = mailfrom[0]
    if recipients is None:
      recipients = util.extractEmails(hdrs.get('to')) or []
      recipients += util.extractEmails(hdrs.get('cc')) or []
      recipients += util.extractEmails(hdrs.get('bcc')) or []
      if not recipients or len(recipients) <= 0:
        raise MissingHeader('email destination ("to") not specified')
    elif isinstance(recipients, basestring):
      recipients = [recipients]
    data = self._getMimeMessage(hdrs)
    if self.manager.modifier:
      mailfrom, recipients, data = self.manager.modifier.modify(
        mailfrom, recipients, data)
    if not isinstance(data, basestring):
      data = data.as_string()
      if not data.endswith('\n'):
        data += '\n'
    self.manager.sender.send(mailfrom, recipients, data)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/10/21
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

# todo: this could be smarter... for example, it could:
#       - detect when references resolve to the same content, but
#         by different Content-IDs
#       - detect when multipart sections could collapse to the same
#         semantic structure

from __future__ import absolute_import
import unittest, email

from .util import smtpHeaderFormat

#------------------------------------------------------------------------------
def canonicalHeaders(message, ignore=None):
  '''
  Returns a canonical string representation of the `message` headers,
  with the following changes made:

  * The MIME boundary specified in the "Content-Type" header, if
    specified, removed.

  * Any headers listed in `ignore` are removed.

  :Parameters:

  ignore : list(str), optional, default: ['Content-Transfer-Encoding']
    List of headers that should not be included in the canonical
    form.
  '''
  if ignore is None:
    ignore = ['Content-Transfer-Encoding']
  ignore = [key.lower() for key in ignore]
  hdrs = {key.lower(): '; '.join(sorted(message.get_all(key)))
          for key in message.keys()
          if key.lower() not in ignore}
  hdrs['content-type'] = '; '.join(['='.join(filter(None, pair))
                                    for pair in message.get_params()
                                    if pair[0].lower() != 'boundary'])
  return '\n'.join([
    smtpHeaderFormat(key) + ': ' + hdrs[key]
    for key in sorted(hdrs.keys())]) + '\n'

#------------------------------------------------------------------------------
def canonicalStructure(message):
  ret = message.get_content_type() + '\n'
  if not message.is_multipart():
    return ret
  msgs = message.get_payload()
  for idx, msg in enumerate(msgs):
    last = idx + 1 >= len(msgs)
    indent = '\n|-- ' if not last else '\n    '
    ret += '|-- ' if not last else '`-- '
    ret += indent.join(canonicalStructure(msg)[:-1].split('\n')) + '\n'
  return ret

#------------------------------------------------------------------------------
def makemsg(msg, submsg):
  if msg is None:
    return submsg
  return msg + ' (' + submsg + ')'

#------------------------------------------------------------------------------
class EmailTestMixin(object):

  mime_cmp_factories = {
    'text/html'      : lambda self, ct: self.try_assertXmlEqual,
    'text/xml'       : lambda self, ct: self.try_assertXmlEqual,
    'text/*'         : lambda self, ct: self.assertMultiLineEqual,
    '*/*'            : lambda self, ct: self.assertEqual,
    }

  #----------------------------------------------------------------------------
  def registerMimeComparator(self, mimetype, comparator):
    def factory(self, ct):
      return comparator
    self.mime_cmp_factories = dict(EmailTestMixin.mime_cmp_factories)
    self.mime_cmp_factories[mimetype] = factory

  #----------------------------------------------------------------------------
  def _parseEmail(self, eml):
    return email.message_from_string(eml)

  #----------------------------------------------------------------------------
  def assertEmailHeadersEqual(self, eml1, eml2, msg=None):
    eml1 = self._parseEmail(eml1)
    eml2 = self._parseEmail(eml2)
    self._assertEmailHeadersEqual(eml1, eml2, msg=msg)

  #----------------------------------------------------------------------------
  def assertNotEmailHeadersEqual(self, eml1, eml2, msg=None):
    try:
      self.assertEmailHeadersEqual(eml1, eml2, msg=msg)
      self.fail(msg or 'email headers %r == %r' % (eml1, eml2))
    except AssertionError: pass

  #----------------------------------------------------------------------------
  def assertEmailStructureEqual(self, eml1, eml2, msg=None):
    eml1 = self._parseEmail(eml1)
    eml2 = self._parseEmail(eml2)
    self._assertEmailStructureEqual(eml1, eml2, msg=msg)

  #----------------------------------------------------------------------------
  def assertNotEmailStructureEqual(self, eml1, eml2, msg=None):
    try:
      self.assertEmailStructureEqual(eml1, eml2, msg=msg)
      self.fail(msg or 'email structure %r == %r' % (eml1, eml2))
    except AssertionError: pass

  #----------------------------------------------------------------------------
  def assertEmailContentEqual(self, eml1, eml2, msg=None, mime_cmp_factories=None):
    eml1 = self._parseEmail(eml1)
    eml2 = self._parseEmail(eml2)
    self._assertEmailContentEqual(eml1, eml2, msg=msg, mcf=mime_cmp_factories)

  #----------------------------------------------------------------------------
  def assertNotEmailContentEqual(self, eml1, eml2, msg=None):
    try:
      self.assertEmailContentEqual(eml1, eml2, msg=msg)
      self.fail(msg or 'email content %r == %r' % (eml1, eml2))
    except AssertionError: pass

  #----------------------------------------------------------------------------
  def assertEmailEqual(self, eml1, eml2, msg=None, mime_cmp_factories=None):
    eml1 = self._parseEmail(eml1)
    eml2 = self._parseEmail(eml2)
    self._assertEmailHeadersEqual(eml1, eml2, msg=msg)
    self._assertEmailStructureEqual(eml1, eml2, msg=msg)
    self._assertEmailContentEqual(eml1, eml2, msg=msg, mcf=mime_cmp_factories)

  #----------------------------------------------------------------------------
  def assertNotEmailEqual(self, eml1, eml2, msg=None, mime_cmp_factories=None):
    try:
      self.assertEmailEqual(eml1, eml2, msg=msg, mime_cmp_factories=mime_cmp_factories)
      self.fail(msg or 'email %r == %r' % (eml1, eml2))
    except AssertionError: pass

  #----------------------------------------------------------------------------
  def _assertEmailHeadersEqual(self, msg1, msg2, msg=None):
    hdr1 = 'EMAIL HEADERS:\n' + canonicalHeaders(msg1)
    hdr2 = 'EMAIL HEADERS:\n' + canonicalHeaders(msg2)
    self.assertMultiLineEqual(hdr1, hdr2, msg=msg)

  #----------------------------------------------------------------------------
  def _assertEmailStructureEqual(self, msg1, msg2, msg=None):
    str1 = 'EMAIL STRUCTURE:\n' + canonicalStructure(msg1)
    str2 = 'EMAIL STRUCTURE:\n' + canonicalStructure(msg2)
    self.assertMultiLineEqual(str1, str2, msg=msg)

  #----------------------------------------------------------------------------
  def _assertEmailContentEqual(self, msg1, msg2, msg=None, mcf=None, context=None):
    if context is None:
      context = 'component root'
    self.assertEqual(
      msg1.is_multipart(), msg2.is_multipart(),
      msg=makemsg(msg, context + ' is not multipart similar'))
    self.assertEqual(
      msg1.get_content_type(), msg2.get_content_type(),
      msg=makemsg(msg, context + ' has content-type mismatch'))
    if context == 'component root':
      context = 'component ' + msg1.get_content_type()
    if not msg1.is_multipart():
      return self._assertEmailPayloadEqual(
        msg1, msg2, msg=msg, mcf=mcf, context=context)
    msgs1 = msg1.get_payload()
    msgs2 = msg2.get_payload()
    self.assertEqual(
      len(msgs1), len(msgs2),
      msg=makemsg(msg, context + ' has sub-message count mismatch'))
    for idx, submsg in enumerate(msgs1):
      sctxt = context + '[' + str(idx) + '] > ' + submsg.get_content_type()
      self._assertEmailContentEqual(
        submsg, msgs2[idx], msg=msg, mcf=mcf, context=sctxt)

  #----------------------------------------------------------------------------
  def _assertEmailPayloadEqual(self, msg1, msg2, msg=None, mcf=None, context='message'):
    # paranoia...
    self.assertFalse(msg1.is_multipart() or msg2.is_multipart())
    self.assertEqual(msg1.get_content_type(), msg2.get_content_type())
    # /paranoia...
    dat1 = msg1.get_payload(decode=True)
    dat2 = msg2.get_payload(decode=True)
    def getcmp(msg, mcf):
      ret = mcf.get(msg.get_content_type())
      if ret is None:
        ret = mcf.get(msg.get_content_maintype() + '/*')
      if ret is None:
        ret = mcf.get('*/*')
      return ret
    pcmp = None
    if mcf is not None:
      pcmp = getcmp(msg1, mcf)
    if pcmp is None:
      pcmp = getcmp(msg1, self.mime_cmp_factories)
    self.assertIsNotNone(
      pcmp, 'no comparator for mime-type "%s"' % (msg1.get_content_type(),))
    pcmp = pcmp(self, msg1.get_content_type())
    try:
      pcmp(dat1, dat2)
    except AssertionError as err:
      raise AssertionError(
        makemsg(msg, context + ' has different payload') + '; ' + err.message)

  #----------------------------------------------------------------------------
  def try_assertXmlEqual(self, dat1, dat2, msg=None):
    if hasattr(self, 'assertXmlEqual'):
      return self.assertXmlEqual(dat1, dat2)
    return self.assertMultiLineEqual(dat1, dat2)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/11/07
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

# TODO: look into:
#   - http://www.ietf.org/rfc/rfc4880.txt
#   - https://www.enigmail.net/documentation/features.php
#   - http://sites.inka.de/tesla/gpgrelay.html

# TODO: add a setting for a key server, and if the recipient is not
#       known, look them up in there... *AWESOME* :)

from __future__ import absolute_import

import logging, asset, email

from .base import Modifier

try:
  import gnupg
  __all__ = ('PgpModifier',)
except ImportError:
  gnupg = None
  __all__ = []

#------------------------------------------------------------------------------

log = logging.getLogger(__name__)

#------------------------------------------------------------------------------
class PgpModifier(Modifier):

  #----------------------------------------------------------------------------
  def __init__(self, prune_keys=True, prune_recipients=False,
               sign=None, add_key='sign-key', gpg_options=None):
    import gnupg
    self.kprune  = prune_keys
    self.rprune  = prune_recipients
    self.sign    = sign
    self.addkeys = add_key
    if self.addkeys is not None:
      if asset.isstr(self.addkeys):
        self.addkeys = [self.addkeys]
      self.addkeys = list(set(self.addkeys))
      if 'sign-key' in self.addkeys:
        self.addkeys.remove('sign-key')
        if self.sign is not None:
          self.addkeys.append(self.sign)
    self.gpg = gnupg.GPG(**(gpg_options or dict()))

  #----------------------------------------------------------------------------
  def modify(self, mailfrom, recipients, data, *other):

    rcptlist = list(set(recipients))

    if self.kprune:
      keys = self.gpg.list_keys()
      for rcpt in rcptlist[:]:
        for key in keys:
          if rcpt in ','.join(key.get('uids', [])):
            break
        else:
          rcptlist.remove(rcpt)
          log.warning('recipient %s removed (not found in keys)', rcpt)

    if self.rprune:
      recipients = rcptlist[:]
      if asset.isstr(data):
        data = email.message_from_string(data)
      if 'to' in data:
        del data['to']
        data.add_header('to', ', '.join(recipients))

    if not asset.isstr(data):
      hdritems = data.items()
      data = data.as_string()
      if not data.endswith('\n'):
        data += '\n'
    else:
      hdritems = email.message_from_string(data).items()

    if self.addkeys:
      rcptlist = list(set(rcptlist + self.addkeys))

    data = self.gpg.encrypt(data, rcptlist, sign=self.sign, always_trust=True)
    if not data.ok:
      raise ValueError('Encryption failed: ' + data.status)

    edata = email.MIMEMultipart.MIMEMultipart(
      'encrypted', protocol='application/pgp-encrypted')

    params = email.MIMENonMultipart.MIMENonMultipart('application', 'pgp-encrypted')
    params.set_payload('Version: 1\n')
    edata.attach(params)

    payload = email.MIMENonMultipart.MIMENonMultipart('application', 'octet-stream')
    payload.set_payload(str(data))
    edata.attach(payload)

    for key, value in hdritems:
      if key.lower() in ('content-type', 'mime-version'):
        continue
      edata.add_header(key, value)

    return (mailfrom, recipients, edata) + other

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/10
# copy: (C) Copyright 2013 Cadit Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import sys, unittest, re
import templatealchemy as ta

from ..manager import Manager
from ..sender import Sender, StoredSender
from .. import util, modifier, testing
from ..test import template, stoptime, unstoptime

#------------------------------------------------------------------------------
class TestBccModifier(unittest.TestCase, testing.EmailTestMixin):

  maxDiff = None

  #----------------------------------------------------------------------------
  def setUp(self):
    stoptime(1234567890.0) # == 2009-02-13T23:31:30Z

  #----------------------------------------------------------------------------
  def tearDown(self):
    unstoptime()

  #----------------------------------------------------------------------------
  def test_bcc(self):
    tpl = 'Hello, {{name}}!'
    manager = Manager(provider = template(tpl, renderer='mustache'),
                      modifier = modifier.BccModifier('bcc@example.com'),
                      sender   = StoredSender(),
                      )
    eml = manager.newEmail()
    eml['name'] = 'Joe Schmoe'
    eml.setHeader('to', 'rcpt@example.com')
    eml.setHeader('from', 'noreply@example.com')
    # override the UNpredictable generated info...
    eml.setHeader('message-id', '<1234567890@@genemail.example.com>')
    eml.includeComponents = ['text']
    eml.send()
    self.assertEqual(1, len(manager.sender.emails))
    out = manager.sender.emails[0]
    self.assertEqual(sorted(out['recipients']), sorted(['rcpt@example.com', 'bcc@example.com']))
    self.assertEqual(out['mailfrom'], 'noreply@example.com')
    self.assertMultiLineEqual(out['message'], '''\
MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: rcpt@example.com
Message-ID: <1234567890@@genemail.example.com>
From: noreply@example.com
Subject: Hello, Joe Schmoe!

Hello, Joe Schmoe!
''')

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  genemail.sender
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/09
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

'''
The low-level mail sending agent. Used by the manager to delegate the
actual sending of the composed email. Note that the primary reason to
use a delegated object for this (rather than just using
smtplib.SMTP()) is so that an email can be serialized into another
form, such as for entry into a database or for unit testing and
comparison.
'''

from __future__ import absolute_import

__all__ = ('Sender', 'StoredSender', 'SmtpSender', 'DebugSender')

import smtplib
import email.parser

from templatealchemy.util import adict

#------------------------------------------------------------------------------
class Sender(object):
  '''
  Abstract interface for an object capable of sending a genemail email
  out, usually to an SMTP MTA.
  '''

  #----------------------------------------------------------------------------
  def send(self, mailfrom, recipients, message):
    '''
    Sends the specified `message` (in SMTP format) to the specified
    `recipients` coming from the email address `mailfrom`. Parameters:

    :Parameters:

    mailfrom : str
      equivalent to the the SMTP ``MAIL FROM`` command.

    recipients : list(str)
      equivalent to the SMTP ``RCPT TO`` command.

    message : str
      the actuall message to be transferred, equivalent to the payload of
      the SMTP ``DATA`` command.
    '''
    raise NotImplementedError()

#------------------------------------------------------------------------------
class SmtpSender(Sender):
  '''
  An implementation of the :class:`genemail.sender.Sender` interface that
  connects to a local or remote SMTP server and submits the message for
  transfer or delivery.

  :Parameters:

  host : str, optional, default: 'localhost'
    the SMTP server to connect to.

  port : int, optional, default: 25
    the SMTP server port to connect to.

  ssl : bool, optional, default: false
    indicates whether or not to connect using SSL.

  starttls : bool, optional, default: false
    indicates that a STARTTLS command should be sent after connecting.

  username : str, optional
    set the SMTP username to authenticate as.

  password : str, optional
    set the password for the `username`.
  '''

  #----------------------------------------------------------------------------
  def __init__(self,
               host='localhost', port=25, ssl=False, starttls=False,
               username=None, password=None, *args, **kwargs):
    super(SmtpSender, self).__init__(*args, **kwargs)
    self.smtpHost = host or 'localhost'
    self.smtpPort = port or 25
    self.username = username
    self.password = password
    self.starttls = starttls
    self.ssl      = ssl

  #----------------------------------------------------------------------------
  def send(self, mailfrom, recipients, message):
    smtp = smtplib.SMTP_SSL() if self.ssl else smtplib.SMTP()
    smtp.connect(self.smtpHost, self.smtpPort)
    if self.starttls:
      smtp.starttls()
    if self.username is not None:
      smtp.login(self.username, self.password)
    smtp.sendmail(mailfrom, recipients, message)
    smtp.quit()

#------------------------------------------------------------------------------
class StoredSender(Sender):
  '''
  An implementation of the :class:`genemail.sender.Sender` interface
  that simply stores all messages in local memory in the
  :attr:`emails` attribute. Most useful when unit testing email
  generation.
  '''

  #----------------------------------------------------------------------------
  def __init__(self, *args, **kwargs):
    super(StoredSender, self).__init__(*args, **kwargs)
    self.emails = []

  #----------------------------------------------------------------------------
  def send(self, mailfrom, recipients, message):
    self.emails.append(
      adict(mailfrom=mailfrom, recipients=recipients, message=message))

#------------------------------------------------------------------------------
class DebugSender(StoredSender):
  '''
  An extension to the :class:`StoredSender` class that parses each
  email into it's MIME components, which simplifies unittesting. Each
  element in the `emails` attribute has the following attributes:

  * `mailfrom`:   SMTP-level `MAIL FROM` value (string)
  * `recipients`: SMTP-level `RCPT TO` value (list)
  * `message`:    raw SMTP `DATA` value (string)
  * `mime`:       the parsed :class:`email.message.Message` object
  * `from`:       email "From" header - not used by SMTP
  * `to`:         email "To" header - not used by SMTP
  * `date`:       email "Date" header
  * `subject`:    email "Subject" header
  * `plain`:      text/plain version of the email (or None)
  * `html`:       text/html version of the email (or None)
  * `calendar`:   text/calendar attachment of the email (or None)
  '''

  #----------------------------------------------------------------------------
  def send(self, mailfrom, recipients, message):
    eml = adict(mailfrom=mailfrom, recipients=recipients, message=message)
    mime = email.parser.Parser().parsestr(message)
    eml['mime']    = mime
    eml['from']    = mime.get('from')
    eml['to']      = mime.get('to')
    eml['date']    = mime.get('date')
    eml['subject'] = mime.get('subject')
    for part in mime.walk():
      ct = part.get_content_type()
      if not ct.startswith('text/'):
        continue
      ct = ct.split('/')[1]
      if eml.get(ct) is None:
        eml[ct] = part.get_payload()
      elif isinstance(eml[ct], list):
        eml[ct].append(part.get_payload())
      else:
        eml[ct] = [eml[ct], part.get_payload()]
    self.emails.append(eml)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# shamelessly scrubbed (and renamed) from:
#   http://code.activestate.com/recipes/66315/
# ==> updated with __contains__

class idict:
  """Dictionary, that has case-insensitive keys.
  
  Keys are retained in their original form
  when queried with .keys() or .items().

  Implementation: An internal dictionary maps lowercase
  keys to (key,value) pairs. All key lookups are done
  against the lowercase keys, but all methods that expose
  keys to the user retrieve the original keys."""
  
  def __init__(self, dict=None):
    """Create an empty dictionary, or update from 'dict'."""
    self._dict = {}
    if dict:
      self.update(dict)

  def __getitem__(self, key):
    """Retrieve the value associated with 'key' (in any case)."""
    k = key.lower()
    return self._dict[k][1]

  def __setitem__(self, key, value):
    """Associate 'value' with 'key'. If 'key' already exists, but
    in different case, it will be replaced."""
    k = key.lower()
    self._dict[k] = (key, value)

  def __delitem__(self, key):
    """Deletes the 'value' associated with 'key'."""
    k = key.lower()
    del self._dict[k]

  def __contains__(self, key):
    """Support for "in" operator."""
    return self.has_key(key)

  def has_key(self, key):
    """Case insensitive test wether 'key' exists."""
    k = key.lower()
    return self._dict.has_key(k)

  def keys(self):
    """List of keys in their original case."""
    return [v[0] for v in self._dict.values()]

  def values(self):
    """List of values."""
    return [v[1] for v in self._dict.values()]

  def items(self):
    """List of (key,value) pairs."""
    return self._dict.values()

  def get(self, key, default=None):
    """Retrieve value associated with 'key' or return default value
    if 'key' doesn't exist."""
    try:
      return self[key]
    except KeyError:
      return default

  def setdefault(self, key, default):
    """If 'key' doesn't exists, associate it with the 'default' value.
    Return value associated with 'key'."""
    if not self.has_key(key):
      self[key] = default
    return self[key]

  def update(self, dict):
    """Copy (key,value) pairs from 'dict'."""
    for k,v in dict.items():
      self[k] = v

  def __repr__(self):
    """String representation of the dictionary."""
    items = ", ".join([("%r: %r" % (k,v)) for k,v in self.items()])
    return "{%s}" % items

  def __str__(self):
    """String representation of the dictionary."""
    return repr(self)



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  genemail.cli
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/09
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import sys, argparse, yaml, os.path, getpass
import templatealchemy as ta
from templatealchemy import stream
from . import manager, sender

#------------------------------------------------------------------------------
def main(args=None, output=None):

  cli = argparse.ArgumentParser(
    description='Command-line interface to the `genemail` email generation'
    ' library.'
    )

  cli.add_argument(
    '-v', '--verbose',
    action='count',
    help='enable verbose output (multiple invocations increase verbosity)')

  cli.add_argument(
    '-n', '--name', metavar='NAME',
    default=None, action='store',
    help='set the template name; if omitted, the root template will be used')

  cli.add_argument(
    '-p', '--param', metavar='NAME=VALUE',
    default=[], action='append',
    help='set a template variable where `VALUE` is taken as a literal'
    ' string (overrides any values set in `--params`)')

  cli.add_argument(
    '-y', '--params', metavar='YAML',
    default=[], action='append',
    help='specifies a YAML-encoded dictionary as template variables;'
    ' if the value starts with the letter "@", the rest is the file name to'
    ' read as a YAML structure. If the value is exactly a dash ("-"), then'
    ' the YAML structure is read from STDIN. In all cases, the YAML'
    ' structure must be a dictionary')

  cli.add_argument(
    '-r', '--renderer', metavar='SPEC',
    default='mako',
    help='sets the TemplateAlchemy rendering driver (default: %(default)r)')

  cli.add_argument(
    '--smtp-host', metavar='HOST',
    default='localhost',
    help='set the SMTP server hostname (default: %(default)r)')

  cli.add_argument(
    '--smtp-port', metavar='PORT',
    default=25, type=int,
    help='set the SMTP server port number (default: %(default)r)')

  cli.add_argument(
    '--smtp-ssl',
    default=False, action='store_true',
    help='enable SSL communication with the SMTP server')

  cli.add_argument(
    '--smtp-starttls',
    default=False, action='store_true',
    help='enable STARTTLS with the SMTP server')

  cli.add_argument(
    '--smtp-username', metavar='USERNAME',
    help='set the SMTP username to authenticate as')

  cli.add_argument(
    '--smtp-password', metavar='PASSWORD',
    help='set the SMTP password to authenticate with (if `--smtp-username`'
    ' is specified, but not `--smtp-password` or the password is exactly'
    ' "-", the password will be prompted for securely)')

  cli.add_argument(
    '-T', '--text',
    action='store_true',
    help='don\'t send the email; just display the TEXT version')

  cli.add_argument(
    '-H', '--html',
    action='store_true',
    help='don\'t send the email; just display the HTML version')

  cli.add_argument(
    '-S', '--standalone',
    action='store_true',
    help='with `--html`, render the "standalone" verion')

  cli.add_argument(
    '-s', '--smtp',
    action='store_true',
    help='don\'t send the email; just display the SMTP request')

  cli.add_argument(
    'source', metavar='SOURCE',
    nargs='?',
    help='sets the TemplateAlchemy source driver; if exactly a dash ("-") or'
    ' omitted, then the template is read from STDIN')

  options = cli.parse_args(args)

  params = dict()

  for yparam in options.params:
    if yparam == '-':
      yparam = sys.stdin.read()
    elif yparam.startswith('@'):
      with open(yparam[1:], 'rb') as fp:
        yparam = fp.read()
    try:
      yparam = yaml.load(yparam)
    except Exception:
      cli.error('could not parse YAML expression: %r'
                % (yparam,))
    if not isinstance(yparam, dict):
      cli.error('"--params" expressions must resolve to dictionaries')
    params.update(yparam)

  for kparam in options.param:
    if '=' not in kparam:
      cli.error('"--param" expressions must be in the KEY=VALUE format')
    key, value = kparam.split('=', 1)
    params[key] = value

  options.params = params
  output = output or sys.stdout

  if options.source is None or options.source == '-':
    options.source = stream.StreamSource(sys.stdin)
  elif ':' not in options.source and os.path.isfile(options.source):
    options.source = stream.StreamSource(open(options.source, 'rb'))

  template = ta.Template(
    source   = options.source,
    renderer = options.renderer,
    )

  if options.smtp_username and (
    not options.smtp_password or options.smtp_password == '-' ):
    options.smtp_password = getpass.getpass(prompt='SMTP Password: ')

  emlsender = sender.SmtpSender(
    host     = options.smtp_host,
    port     = options.smtp_port,
    ssl      = options.smtp_ssl,
    starttls = options.smtp_starttls,
    username = options.smtp_username,
    password = options.smtp_password,
    )

  emlman = manager.Manager(provider=template, sender=emlsender)
  eml    = emlman.newEmail(options.name)

  for k, v in options.params.items():
    eml[k] = v

  if options.text:
    sys.stdout.write(eml.getText())
    return 0

  if options.html:
    sys.stdout.write(eml.getHtml(standalone=options.standalone))
    return 0

  if options.smtp:
    sys.stdout.write(eml.getSmtpData())
    return 0

  eml.send()
  return 0

#------------------------------------------------------------------------------
if __name__ == '__main__':
  sys.exit(main())

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/31
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

from __future__ import absolute_import

from .base import Modifier

try:
  import dkim
  __all__ = ('DkimModifier',)
except ImportError:
  dkim = None
  __all__ = []

#------------------------------------------------------------------------------
class DkimModifier(Modifier):

  #----------------------------------------------------------------------------
  def __init__(self, privateKey, domain, selector,
               identity=None, canonicalize=None, headers=None,
               algorithm=None, length=None):
    '''
    Create a genemail.Modifier that adds DKIM signature headers to emails.

    :Parameters:

    privateKey : str
      a PKCS#1 private key in base64-encoded text form

    domain : str
      the DKIM domain value for the signature

    selector : str
      the DKIM selector value for the signature

    identity : str, optional
      the DKIM identity value for the signature (defaults to the
      default dkimpy package value)

    canonicalize : list(str, str), optional
      the canonicalization algorithms to use (defaults to the default
      dkimpy package value)

    headers : list(str), optional
      a list of strings indicating which headers are to be signed
      (defaults to the default dkimpy package value)

    algorithm : str, optional
      a list of strings indicating which headers are to be signed
      (defaults to the default dkimpy package value)

    length : bool, optional
      true if the l= tag should be included to indicate body length
      (defaults to the default dkimpy package value)
    '''
    import dkim
    self.key      = privateKey
    self.domain   = domain
    self.selector = selector
    self.logger   = None
    self.dkimargs = dict()
    if identity is not None:
      self.dkimargs['identity'] = identity
    if canonicalize is not None:
      self.dkimargs['canonicalize'] = canonicalize
    if headers is not None:
      self.dkimargs['include_headers'] = headers
    if algorithm is not None:
      self.dkimargs['signature_algorithm'] = algorithm
    if length is not None:
      self.dkimargs['length'] = length

  #----------------------------------------------------------------------------
  def modify(self, mailfrom, recipients, data):
    import dkim
    if not isinstance(data, basestring):
      data = data.as_string()
    dkimhdr = dkim.sign(data, self.selector, self.domain, self.key,
                        logger=self.logger, **self.dkimargs)
    return (mailfrom, recipients, dkimhdr + data)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/11/07
# copy: (C) Copyright 2013 Cadit Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import os, sys, unittest, re
import templatealchemy as ta

from .. import testing
from ..manager import Manager
from ..sender import StoredSender
from ..test import template, stoptime, unstoptime, extrafeature
from .pgp import PgpModifier

moddir = os.path.dirname(os.path.dirname(__file__))

#------------------------------------------------------------------------------
class TestPgpModifier(unittest.TestCase, testing.EmailTestMixin):

  maxDiff = None

  encEmail = '''\
MIME-Version: 1.0
Content-Type: multipart/encrypted; protocol=application/pgp-encrypted; boundary="==BNDRY=="
Content-Transfer-Encoding: 7bit
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: receiver@example.com, the-nsa@example.com
Message-ID: <1234567890@genemail.example.com>
From: sender@example.com
Subject: Hello, Joe Schmoe!

--==BNDRY==
MIME-Version: 1.0
Content-Type: application/pgp-encrypted

Version: 1

--==BNDRY==
Content-Type: application/octet-stream
MIME-Version: 1.0

-----BEGIN PGP MESSAGE-----
Version: GnuPG v1.4.11 (GNU/Linux)

hIwDPf0Efz9KxdABBACt1HKG6E+ONo1CvvH9fAbgyeEBI0yW7IqSIXewU7Il3odJ
fmgX5eQUP/HE/Jf+XsLaLZ0IZq/Jcn5hD1JzNdDStc9Kk9ry3Dwq3WTwLy4Gq6Ji
QEqYKxvUEx/VFqHfo68XiFJTDl9HrIXMeIoV0oQrirF1xkXFo88LuoPGUmAixNLA
UQHrgiKqP142RWT1KUXJzgIFhM95em25wkcdhLuAsJJG/a4fwT9gzL0hhCMnmwce
yHdCpen7umdclaHvoZKNczrhPNLRt/X3IJlyztqTsJS4iavRO+usQRP/KNjtxlZE
m9j7uLcGH8lUXgnd4hChFWojV0BLpR6tRIaGLPWeopx331BBTz9CzeIjgJGJnYvy
8aBqjQoNiq1aEYM9DQX/QjsCqjlooPvBHJNKVbeTUszPQIh+wFqj0sUcXLsMwq4F
bUFQLIiSDzveKLsDJVt+2iAvaSNs/WAjtunmCd4qXKb4F0RXKPit+0Gks3XLDB6L
vJdCbGtFxl8NNYfL4g4FiCTMt7GgumGw7kz4C6V7CTaoDQ==
=ip5F
-----END PGP MESSAGE-----

--==BNDRY==--
'''

  #----------------------------------------------------------------------------
  def setUp(self):
    stoptime(1234567890.0) # == 2009-02-13T23:31:30Z 

  #----------------------------------------------------------------------------
  def tearDown(self):
    unstoptime()

  #----------------------------------------------------------------------------
  @extrafeature('pgp')
  def test_defaults(self):
    tpl = 'Hello, {{name}}!'
    pgpmod = PgpModifier(
      gpg_options=dict(
        gnupghome=os.path.join(moddir, 'test_data', 'gpg-sender')
        )
      )
    manager = Manager(provider = template(tpl, renderer='mustache'),
                      modifier = pgpmod,
                      sender   = StoredSender(),
                      )
    eml = manager.newEmail()
    eml['name'] = 'Joe Schmoe'
    eml.setHeader('to', 'receiver@example.com, the-nsa@example.com')
    eml.setHeader('from', 'sender@example.com')
    # override the UNpredictable generated info...
    eml.setHeader('message-id', '<1234567890@genemail.example.com>')
    eml.includeComponents = ['text']
    eml.send()
    self.assertEqual(1, len(manager.sender.emails))
    out = manager.sender.emails[0]
    chk = self.encEmail
    self.registerMimeComparator('application/octet-stream', self.assertEncryptedEmailEqual_defaults)
    self.assertEmailEqual(out['message'], chk)
    self.assertEqual(out['recipients'], ['receiver@example.com', 'the-nsa@example.com'])

    # TODO: validate that the email is:
    #   - signed
    #   - encrypted to both receiver and sender
    #   - not encrypted for anyone else

  #----------------------------------------------------------------------------
  def assertEncryptedEmailEqual_defaults(self, a, b, msg=None):
    import gnupg
    gpg = gnupg.GPG(gnupghome=os.path.join(moddir, 'test_data', 'gpg-receiver'))
    adec = gpg.decrypt(a)
    if not adec.ok:
      self.fail('could not decrypt message: ' + a)
    bdec = gpg.decrypt(b)
    if not bdec.ok:
      self.fail('could not decrypt message: ' + b)
    self.assertEmailEqual(str(adec), str(bdec), msg=msg)

  #----------------------------------------------------------------------------
  @extrafeature('pgp')
  def test_prune_recipients(self):
    tpl = 'Hello, {{name}}!'
    pgpmod = PgpModifier(
      prune_keys=True,
      prune_recipients=True,
      gpg_options=dict(
        gnupghome=os.path.join(moddir, 'test_data', 'gpg-sender')
        )
      )
    manager = Manager(provider = template(tpl, renderer='mustache'),
                      modifier = pgpmod,
                      sender   = StoredSender(),
                      )
    eml = manager.newEmail()
    eml['name'] = 'Joe Schmoe'
    eml.setHeader('to', 'receiver@example.com, the-nsa@example.com')
    eml.setHeader('from', 'sender@example.com')
    # override the UNpredictable generated info...
    eml.setHeader('message-id', '<1234567890@genemail.example.com>')
    eml.includeComponents = ['text']
    eml.send()
    self.assertEqual(1, len(manager.sender.emails))
    out = manager.sender.emails[0]
    msg = out['message']
    chk = self.encEmail.replace(', the-nsa@example.com', '')
    self.registerMimeComparator('application/octet-stream', self.assertEncryptedEmailEqual_prune)
    self.assertEmailEqual(out['message'], chk)
    self.assertEqual(out['recipients'], ['receiver@example.com'])

  #----------------------------------------------------------------------------
  def assertEncryptedEmailEqual_prune(self, a, b, msg=None):
    import gnupg
    gpg = gnupg.GPG(gnupghome=os.path.join(moddir, 'test_data', 'gpg-receiver'))
    adec = gpg.decrypt(a)
    if not adec.ok:
      self.fail('could not decrypt message: ' + a)
    chk = '''\
MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: receiver@example.com
Message-ID: <1234567890@genemail.example.com>
From: sender@example.com
Subject: Hello, Joe Schmoe!

Hello, Joe Schmoe!
'''
    self.assertEmailEqual(str(adec), chk, msg=msg)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/10/22
# copy: (C) Copyright 2013 Cadit Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import sys, unittest
try:
  import pxml
except ImportError:
  pxml = None

from .testing import EmailTestMixin

#------------------------------------------------------------------------------
class TestEmailMixin(EmailTestMixin, unittest.TestCase):

  maxDiff = None

  #----------------------------------------------------------------------------
  def setUp(self):
    super(TestEmailMixin, self).setUp()
    self.noxml = False

  #----------------------------------------------------------------------------
  def test_mixin_headers_same(self):
    eml1 = '''\
Content-Type: multipart/alternative; boundary="==genemail.test-alt-2=="
MIME-Version: 1.0
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: test@example.com
Message-ID: <1234567890@@genemail.example.com>
From: noreply@example.com
Subject: Foo The Bar

--==genemail.test-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

CONTENT
--==genemail.test-alt-2==
'''
    eml2 = '''\
date: Fri, 13 Feb 2009 23:31:30 -0000
subject: Foo The Bar
from: noreply@example.com
mime-version: 1.0
to: test@example.com
content-type: multipart/alternative; boundary="==genemail.test-BOUNDARY-alt-2=="
message-id: <1234567890@@genemail.example.com>

--==genemail.test-BOUNDARY-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

CONTENT
--==genemail.test-BOUNDARY-alt-2==
'''
    self.assertEmailEqual(eml1, eml2)

  #----------------------------------------------------------------------------
  def test_mixin_headers_diff(self):
    eml1 = '''\
Content-Type: text/plain
MIME-Version: 1.0
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: test@example.com
Message-ID: <1234567890@@genemail.example.com>
From: noreply@example.com
Subject: Foo The Bar

CONTENT
'''
    eml2 = '''\
Date: Fri, 13 Feb 2009 23:31:30 -0000
Subject: Foo The Bar
X-Generator: an extra header (note that mime-version is missing)
From: noreply@example.com
To: test@example.com
Message-ID: <1234567890@@genemail.example.com>
Content-Type: text/plain; charset=us-ascii

CONTENT
'''
    with self.assertRaises(AssertionError) as cm:
      self.assertEmailEqual(eml1, eml2)
    msg = '\n'.join(cm.exception.message.split('\n')[1:])
    self.assertMultiLineEqual(msg, '''\
  EMAIL HEADERS:
- Content-Type: text/plain
+ Content-Type: text/plain; charset=us-ascii
  Date: Fri, 13 Feb 2009 23:31:30 -0000
  From: noreply@example.com
  Message-ID: <1234567890@@genemail.example.com>
- MIME-Version: 1.0
  Subject: Foo The Bar
  To: test@example.com
+ X-Generator: an extra header (note that mime-version is missing)
''')

  #----------------------------------------------------------------------------
  def test_mixin_structure_same(self):
    eml1 = '''\
Content-Type: multipart/alternative; boundary="==genemail.test-alt-2=="
MIME-Version: 1.0

--==genemail.test-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

--==genemail.test-alt-2==
Content-Type: multipart/related; boundary="==genemail.test-rel-3=="
MIME-Version: 1.0

--==genemail.test-rel-3==
Content-Type: text/plain
MIME-Version: 1.0

--==genemail.test-rel-3==
Content-Type: image/png
MIME-Version: 1.0

--==genemail.test-rel-3==--
--==genemail.test-alt-2==--
'''
    eml2 = '''\
Content-Type: multipart/alternative; boundary="==BOUNDARY-f8967b6d-alt-2=="
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-alt-2==
Content-Type: multipart/related; boundary="==BOUNDARY-f8967b6d-rel-3=="
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-rel-3==
Content-Type: text/plain
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-rel-3==
Content-Type: image/png
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-rel-3==--
--==BOUNDARY-f8967b6d-alt-2==--
'''
    self.assertEmailEqual(eml1, eml2)

  #----------------------------------------------------------------------------
  def test_mixin_structure_diff(self):
    eml1 = '''\
Content-Type: multipart/alternative; boundary="==genemail.test-alt-2=="
MIME-Version: 1.0

--==genemail.test-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

--==genemail.test-alt-2==
Content-Type: multipart/related; boundary="==genemail.test-rel-3=="
MIME-Version: 1.0

--==genemail.test-rel-3==
Content-Type: text/plain
MIME-Version: 1.0

--==genemail.test-rel-3==
Content-Type: image/png
MIME-Version: 1.0

--==genemail.test-rel-3==--
--==genemail.test-alt-2==--
'''
    eml2 = '''\
Content-Type: multipart/alternative; boundary="==BOUNDARY-f8967b6d-alt-2=="
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-alt-2==
Content-Type: multipart/related; boundary="==BOUNDARY-f8967b6d-rel-3=="
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-rel-3==
Content-Type: text/plain
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-rel-3==
Content-Type: image/png
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-rel-3==
Content-Type: text/svg
MIME-Version: 1.0

--==BOUNDARY-f8967b6d-rel-3==--
--==BOUNDARY-f8967b6d-alt-2==--
'''
    with self.assertRaises(AssertionError) as cm:
      self.assertEmailEqual(eml1, eml2)
    msg = '\n'.join(cm.exception.message.split('\n')[1:])
    self.assertMultiLineEqual(msg, '''\
  EMAIL STRUCTURE:
  multipart/alternative
  |-- text/plain
  `-- multipart/related
      |-- text/plain
-     `-- image/png
?     ^
+     |-- image/png
?     ^
+     `-- text/svg
''')

  #----------------------------------------------------------------------------
  def test_mixin_content_textplain_same(self):
    eml1 = '''\
Content-Type: text/plain
MIME-Version: 1.0

this is some content.
'''
    eml2 = '''\
Content-Type: text/plain
Content-Transfer-Encoding: quoted-printable
MIME-Version: 1.0

this is some=
 content.
'''
    self.assertEmailEqual(eml1, eml2)

  #----------------------------------------------------------------------------
  def test_mixin_content_textplain_diff(self):
    eml1 = '''\
Content-Type: text/plain
MIME-Version: 1.0

this is some content.
'''
    eml2 = '''\
Content-Type: text/plain
MIME-Version: 1.0

this is some=
 content.
'''
    with self.assertRaises(AssertionError) as cm:
      self.assertEmailEqual(eml1, eml2)
    msg = '\n'.join(cm.exception.message.split('\n')[1:])
    self.assertMultiLineEqual(msg, '''\
- this is some content.
+ this is some=
+  content.
''')

  #----------------------------------------------------------------------------
  def test_mixin_content_multiparttextplain_same(self):
    eml1 = '''\
Content-Type: multipart/alternative; boundary="==genemail.test-alt-2=="
MIME-Version: 1.0

--==genemail.test-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

this is some content.
--==genemail.test-alt-2==--
'''
    eml2 = '''\
Content-Type: multipart/alternative; boundary="==genemail.test-alt-2=="
MIME-Version: 1.0

--==genemail.test-alt-2==
Content-Type: text/plain
Content-Transfer-Encoding: quoted-printable
MIME-Version: 1.0

this is some con=
tent.
--==genemail.test-alt-2==--
'''
    self.assertEmailEqual(eml1, eml2)

  #----------------------------------------------------------------------------
  def test_mixin_content_multiparttextplain_diff(self):
    eml1 = '''\
Content-Type: multipart/alternative; boundary="==genemail.test-alt-2=="
MIME-Version: 1.0

--==genemail.test-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

this is some content.
--==genemail.test-alt-2==--
'''
    eml2 = '''\
Content-Type: multipart/alternative; boundary="==genemail.test-alt-2=="
MIME-Version: 1.0

--==genemail.test-alt-2==
Content-Type: text/plain
MIME-Version: 1.0

this is some con=
tent.
--==genemail.test-alt-2==--
'''
    with self.assertRaises(AssertionError) as cm:
      self.assertEmailEqual(eml1, eml2)
    msg = '\n'.join(cm.exception.message.split('\n')[1:])
    self.assertMultiLineEqual(msg, '''\
- this is some content.
+ this is some con=
tent.
?                 ++
''')

  #----------------------------------------------------------------------------
  def assertXmlEqual(self, x1, x2, msg=None):
    if self.noxml or pxml is None:
      return self.assertMultiLineEqual(x1, x2, msg=msg)
    class PxmlXmlTest(pxml.XmlTestMixin, unittest.TestCase):
      def runTest(self): pass
    PxmlXmlTest().assertXmlEqual(x1, x2, msg=msg)

  #----------------------------------------------------------------------------
  def test_mixin_content_texthtml_same(self):
    eml1 = '''\
Content-Type: text/html
MIME-Version: 1.0

<html><body id="Foo" class="bar">hello</body></html>
'''
    eml2 = '''\
Content-Type: text/html
Content-Transfer-Encoding: quoted-printable
MIME-Version: 1.0

<html  ><body class='bar' id='Foo' >hello=
</body  ></html>
'''
    # note: these are not actually semantically different, but
    #       this is a test of behaviour if 'assertXmlEqual'
    #       is NOT available.
    self.noxml = True
    with self.assertRaises(AssertionError) as cm:
      self.assertEmailEqual(eml1, eml2)
    msg = '\n'.join(cm.exception.message.split('\n')[1:])
    self.assertMultiLineEqual(msg, '''\
- <html><body id="Foo" class="bar">hello</body></html>
?             ---------      ^   ^
+ <html  ><body class='bar' id='Foo' >hello</body  ></html>
?      ++             ^   ^^^^^^^^^^^            ++
''')
    if pxml is None:
      sys.stderr.write('*** PXML LIBRARY NOT PRESENT - SKIPPING XML DIFF *** ')
      return
    self.noxml = False
    self.assertEmailEqual(eml1, eml2)

  #----------------------------------------------------------------------------
  def test_mixin_content_texthtml_diff(self):
    eml1 = '''\
Content-Type: text/html
MIME-Version: 1.0

<html><body id="Foo" class="bar">hello</body></html>
'''
    eml2 = '''\
Content-Type: text/html
Content-Transfer-Encoding: quoted-printable
MIME-Version: 1.0

<html ><body class = 'bar' >hel=
lo</body></html>
'''
    # note: these are both syntactically AND semantically
    # different... they should be different with and without
    # xml processing - but the errors should be different.
    self.noxml = True
    with self.assertRaises(AssertionError) as cm:
      self.assertEmailEqual(eml1, eml2)
    msg = '\n'.join(cm.exception.message.split('\n')[1:])
    self.assertMultiLineEqual(msg, '''\
- <html><body id="Foo" class="bar">hello</body></html>
?             ---------      ^   ^
+ <html ><body class = 'bar' >hello</body></html>
?      +            + ^^   ^^
''')
    if pxml is None:
      sys.stderr.write('*** PXML LIBRARY NOT PRESENT - SKIPPING XML DIFF *** ')
      return
    self.noxml = False
    with self.assertRaises(AssertionError) as cm:
      self.assertEmailEqual(eml1, eml2)
    msg = '\n'.join(cm.exception.message.split('\n')[1:])
    self.assertMultiLineEqual(msg, '''\
  <?xml version="1.0" encoding="UTF-8"?>
  <html>
-   <body class="bar" id="Foo">hello</body>
?                   ---------
+   <body class="bar">hello</body>
  </html>
''')

  #----------------------------------------------------------------------------
  def test_mixin_allinone(self):

    if pxml is None:
      sys.stderr.write('*** PXML LIBRARY NOT PRESENT - SKIPPING XML DIFF *** ')
      return

    eml1 = '''\
Content-Type: multipart/alternative; boundary="==genemail.test-alt-2=="
MIME-Version: 1.0
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: test@example.com
Message-ID: <1234567890@@genemail.example.com>
From: noreply@example.com
Subject: Foo The Bar

--==genemail.test-alt-2==
MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit

Foo the bar [].

--==genemail.test-alt-2==
Content-Type: multipart/related; boundary="==genemail.test-rel-3=="
MIME-Version: 1.0

--==genemail.test-rel-3==
MIME-Version: 1.0
Content-Type: text/html; charset="us-ascii"
Content-Transfer-Encoding: 7bit

<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>Foo The Bar</title>
  </head>
  <body id="bar" class="foo">
    <p>Foo the bar <img src="cid:slogan.txt" />.</p>
  </body>
</html>
--==genemail.test-rel-3==
Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Disposition: attachment
Content-ID: <slogan.txt>

ALL YOUR BASE ARE BELONG TO US
--==genemail.test-rel-3==--
--==genemail.test-alt-2==--
'''

    eml2 = '''\
Content-Type: multipart/alternative; boundary="==ARANDOMBOUNDARY-HEHE-alt-2=="
MIME-Version: 1.0
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: test@example.com
Message-ID: <1234567890@@genemail.example.com>
From: noreply@example.com
Subject: Foo The Bar

--==ARANDOMBOUNDARY-HEHE-alt-2==
MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit

Foo the bar [].

--==ARANDOMBOUNDARY-HEHE-alt-2==
Content-Type: multipart/related; boundary="==ARANDOMBOUNDARY-HEHE-rel-3=="
MIME-Version: 1.0

--==ARANDOMBOUNDARY-HEHE-rel-3==
MIME-Version: 1.0
Content-Type: text/html; charset="us-ascii"
Content-Transfer-Encoding: 7bit

<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>Foo The Bar</title>
  </head>
  <body class="foo" id="bar">
    <p>Foo the bar <img src="cid:slogan.txt" />.</p>
  </body>
</html>
--==ARANDOMBOUNDARY-HEHE-rel-3==
Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Disposition: attachment
Content-ID: <slogan.txt>

ALL YOUR BASE ARE BELONG TO US
--==ARANDOMBOUNDARY-HEHE-rel-3==--
--==ARANDOMBOUNDARY-HEHE-alt-2==--
'''
    self.assertEmailEqual(eml1, eml2)


#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/10
# copy: (C) Copyright 2013 Cadit Inc., All Rights Reserved.
#------------------------------------------------------------------------------

from __future__ import absolute_import

import sys, unittest, re
import templatealchemy as ta

from ..manager import Manager
from ..sender import StoredSender
from .. import util, modifier, testing
from ..test import template, stoptime, unstoptime, extrafeature

#------------------------------------------------------------------------------
class TestDkimModifier(unittest.TestCase, testing.EmailTestMixin):

  maxDiff = None

  #----------------------------------------------------------------------------
  def setUp(self):
    stoptime(1234567890.0) # == 2009-02-13T23:31:30Z

  #----------------------------------------------------------------------------
  def tearDown(self):
    unstoptime()

  #----------------------------------------------------------------------------
  @extrafeature('dkim')
  def test_dkim(self):
    tpl = 'Hello, {{name}}!'
    dkimmod = modifier.DkimModifier(
      privateKey = '''\
-----BEGIN RSA PRIVATE KEY-----
MIIBOgIBAAJBAM3tYc678cHyJdmBCoQyqIHv9+eCksiAvc19zNlaDMkVKtk2/yae
8LLSAyf4B4CG1c18HyAv0lS6UeGDzLqk5bMCAwEAAQJAMLp/boAi0RYPxsw2RNoH
7ddu/iVzvmZYg4vFMZmRdPNUaDgJSuPGR1CawTeXwTP+IZuMQHLVDEiVN7g4KSIp
oQIhAPYEF/OWK93cEqGW9fdHlvMwRLekuVciXkypoSESKpZvAiEA1kjMnT7AkvPz
k7gw/rYRo0O3q7EYOylrHde4rVpCZv0CIFp10uDMnUCtBWTJf5P3jPfLDdmBBm2V
w5ro3MiuR16dAiEAp5SVWKA70HEyW6MfxgMzdgA+gupjrdjtaZBMYF4HMi0CIDcs
ZZ1QMmhqF106XrQeSToj+xD2C8INmZh5HyzQJi6h
-----END RSA PRIVATE KEY-----
''',
      domain = 'example.com',
      selector = '20130731v1',
      )
    manager = Manager(provider = template(tpl, renderer='mustache'),
                      modifier = dkimmod,
                      sender   = StoredSender(),
                      )
    eml = manager.newEmail()
    eml['name'] = 'Joe Schmoe'
    eml.setHeader('to', 'rcpt@example.com')
    eml.setHeader('from', 'noreply@example.com')
    # override the UNpredictable generated info...
    eml.setHeader('message-id', '<1234567890@@genemail.example.com>')
    eml.includeComponents = ['text']
    eml.send()
    self.assertEqual(1, len(manager.sender.emails))
    out = manager.sender.emails[0]

    self.assertRegexpMatches(out['message'], '''\
^DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/simple; d=example\.com; i=@example\.com; \r
 q=dns/txt; s=20130731v1; t=\d{1,11}; h=MIME-Version : Content-Type : \r
 Content-Transfer-Encoding : Date : To : Message-ID : From : Subject : \r
 Date : From : Subject; bh=J5dIdGf2WyNWDpPJz50a\+Olx8RJYoHCRY7ckDZ4gT84=; \r
 b=[0-9a-zA-Z=+/\r\n ]{85,}\r
MIME-Version: 1\.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit
Date: Fri, 13 Feb 2009 23:31:30 -0000
To: rcpt@example\.com
Message-ID: <1234567890@@genemail\.example\.com>
From: noreply@example\.com
Subject: Hello, Joe Schmoe!

Hello, Joe Schmoe!$''')

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# lib:  genemail.manager
# auth: Philip J Grabner <grabner@cadit.com>
# date: 2013/07/09
# copy: (C) Copyright 2013 Cadit Health Inc., All Rights Reserved.
#------------------------------------------------------------------------------

from __future__ import absolute_import

__all__ = ('Manager',)

import templatealchemy as TA
from templatealchemy.util import adict, callingPkgName
from . import sender as sendermod
from . import email

#------------------------------------------------------------------------------
class Manager(object):

  #----------------------------------------------------------------------------
  def __init__(self, provider=None, modifier=None, sender=None, default=None):
    '''
    A ``genemail.manager.Manager`` object is the main clearinghouse
    for generating templatized emails. The main objective is that it
    provides access to :class:`genemail.email.Email` objects.

    :Parameters:

    provider : TemplateAlchemy.Template, optional
      the template provider; if not specified, defaults to loading a
      `pkg` provider in the calling package's namespace.

    modifier : :class:`genemail.modifier.Modifier`, optional
      an email modifier, generally used for post-generation email
      adjustment, such as always adding a blind-copy recipient
      (equivalent to adding a BCC header) or adding cryptographic
      signature headers or content with DKIM or GPG.

    sender : :class:`genemail.sender.Sender`, optional
      the email sending agent; if not specified, defaults to a
      SmtpSender with default parameters.

    default : { dict, :class:`genemail.email.Email` }, optional
      sets the default Email object whose attributes are deep-copied
      into a new Email object (created with :meth:`newEmail`).
    '''
    self.provider = provider
    if not self.provider:
      self.provider = TA.Template(
        source='pkg:%s:' % (callingPkgName(ignore='genemail'),))
    self.modifier = modifier or None
    self.sender   = sender or sendermod.SmtpSender()
    self.default  = default
    if not isinstance(self.default, email.Email):
      self.default = email.Email(self, None)
      for attr, val in (default or {}).items():
        if attr in email.Email.DEFAULTS:
          setattr(self.default, attr, val)

  #----------------------------------------------------------------------------
  def newEmail(self, name=None, provider=None, default=None):
    return email.Email(
      self, name, provider=provider, default=default or self.default)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------
